/*
 *	wiiuse
 *
 *	Written By:
 *		Michael Laforest	< para >
 *		Email: < thepara (--AT--) g m a i l [--DOT--] com >
 *
 *	Copyright 2005
 *
 *	This file is part of wiiuse.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	$Header$
 *
 */

/**
 *	@file
 *	@brief General internal wiiuse stuff.
 *
 *	Since Wiiuse is a library, wiiuse.h is a duplicate
 *	of the API header.
 *
 *	The code that would normally go in that file, but
 *	which is not needed by third party developers,
 *	is put here.
 *
 *	So wiiuse_internal.h is included by other files
 *	internally, wiiuse.h is included only here.
 */

#ifndef WIIUSE_INTERNAL_H_INCLUDED
#define WIIUSE_INTERNAL_H_INCLUDED

#include <arpa/inet.h>				/* htons() */
#include <bluetooth/bluetooth.h>
#include "definitions.h"

/********************
 *
 *	Wiimote internal codes
 *
 ********************/

/* Communication channels */
#define WIIMOTE_OUTPUT_CHANNEL			0x11
#define WIIMOTE_INPUT_CHANNEL			0x13

#define SET_REPORT						0x50

/* commands */
#define WIIMOTE_CMD_LED					0x11
#define WIIMOTE_CMD_REPORT_TYPE			0x12
#define WIIMOTE_CMD_RUMBLE				0x13
#define WIIMOTE_CMD_CTRL_STATUS			0x15
#define WIIMOTE_CMD_READ_DATA			0x17
#define WIIMOTE_CMD_WRITE_DATA			0x16

/* input report ids */
#define WIIMOTE_REPORT_CTRL_STATUS		0x20
#define WIIMOTE_REPORT_READ				0x21
#define WIIMOTE_REPORT_WRITE			0x22
#define WIIMOTE_REPORT_BTN				0x30
#define WIIMOTE_REPORT_BTN_MOTION		0x31
#define WIIMOTE_REPORT_BTN_EXP			0x36
#define WIIMOTE_REPORT_BTN_MOTION_EXP	0x37

#define BT_INPUT						0x01
#define BT_OUTPUT						0x02

/* Identify the wiimote device by its class */
#define WIIMOTE_DEV_CLASS_0				0x04
#define WIIMOTE_DEV_CLASS_1				0x25
#define WIIMOTE_DEV_CLASS_2				0x00

/* controller status stuff */
#define WIIMOTE_MAX_BATTERY_CODE		0xC8

/* offsets in wiimote memory */
#define WIIMOTE_MEM_OFFSET_CALIBRATION	0x16

#define WIIMOTE_EXTENSION_MEM_BASE		0x04A40000
#define WIIMOTE_EXTENSION_MEM_ENABLE	0x04A40040
#define WIIMOTE_EXTENSION_MEM_CALIBR	0x04A40020

/* controller status flags for the first message byte */
/* bit 1 is unknown */
#define WIIMOTE_CTRL_STATUS_BYTE1_ATTACHMENT		0x02
#define WIIMOTE_CTRL_STATUS_BYTE1_SPEAKER_ENABLED	0x04
#define WIIMOTE_CTRL_STATUS_BYTE1_IR_ENABLED		0x08
#define WIIMOTE_CTRL_STATUS_BYTE1_LED_1				0x10
#define WIIMOTE_CTRL_STATUS_BYTE1_LED_2				0x20
#define WIIMOTE_CTRL_STATUS_BYTE1_LED_3				0x40
#define WIIMOTE_CTRL_STATUS_BYTE1_LED_4				0x80

/********************
 *
 *	End Wiimote internal codes
 *
 ********************/

/* wiimote state flags */
#define WIIMOTE_INIT_STATES				0x00
#define WIIMOTE_STATE_DEV_FOUND			0x01
#define WIIMOTE_STATE_HANDSHAKE			0x02	/* actual connection exists but no handshake yet */
#define WIIMOTE_STATE_CONNECTED			0x04
#define WIIMOTE_STATE_RUMBLE			0x08
#define WIIMOTE_STATE_MOTION_SENSING	0x10
#define WIIMOTE_STATE_USE_EXPANSION		0x20

/* wiimote extension flags */
#define WIIMOTE_EXT_STATE_HANDSHAKE		0x00	/* actual connection exists but no handshake yet */
#define WIIMOTE_EXT_STATE_CONNECTED		0x01

/* macro to manage states */
#define WIIMOTE_IS_SET(wm, s)			((wm->state & (s)) == (s))
#define WIIMOTE_ENABLE_STATE(wm, s)		(wm->state |= (s))
#define WIIMOTE_DISABLE_STATE(wm, s)	(wm->state &= ~(s))
#define WIIMOTE_TOGGLE_STATE(wm, s)		((wm->state & (s)) ? WIIMOTE_DISABLE_STATE(wm, s) : WIIMOTE_ENABLE_STATE(wm, s))

/* misc macros */
#define WIIMOTE_ID(wm)					(wm->unid)
#define WIIMOTE_IS_CONNECTED(wm)		(WIIMOTE_IS_SET(wm, WIIMOTE_STATE_CONNECTED))

/* convert to big endian */
#define BIG_ENDIAN_LONG(i)				(htonl(i))
#define BIG_ENDIAN_SHORT(i)				(htons(i))


#include "wiiuse.h"


/* Error codes strings */
struct errno_str_t {
	enum WIIUSE_ERROR_CODE code;
	char* str;
};


#ifdef __cplusplus
extern "C" {
#endif

struct wiimote_t** wiimote_init(int wiimotes, int* unids, wiimote_event_cb event_cb, wiimote_ctrl_status_cb stat_cb, wiimote_dis_cb dis_cb);

void wiimote_disconnected(struct wiimote_t* wm);

void wiimote_rumble(struct wiimote_t* wm, int status);
void wiimote_toggle_rumble(struct wiimote_t* wm);

void wiimote_set_leds(struct wiimote_t* wm, int leds);
void wiimote_motion_sensing(struct wiimote_t* wm, int status);

void wiimote_set_report_type(struct wiimote_t* wm);

int wiimote_read_data(struct wiimote_t* wm, wiimote_read_cb read_cb, byte* buffer, unsigned int offset, unsigned short len);

void wiimote_status(struct wiimote_t* wm);

struct wiimote_t* wiimote_get_by_id(struct wiimote_t** wm, int wiimotes, int unid);

int wiimote_write_data(struct wiimote_t* wm, unsigned int addr, byte* data, byte len);

/* not part of the api */
void wiimote_send_next_pending_read_request(struct wiimote_t* wm);
void wiimote_send_report_type(struct wiimote_t* wm);

void wiimote_send(struct wiimote_t* wm, int report_type, byte* msg, int len);

const char* wiimote_get_errno_str(struct wiimote_t* wm);

#ifdef __cplusplus
}
#endif

#endif /* WIIUSE_INTERNAL_H_INCLUDED */
